# Gitを用いた開発の概要

## リポジトリのセットアップ

開発を始める前に、まずリモートリポジトリのローカルコピーを作成する必要があります。これをクローンと呼びます。

### クローン

HTTPSまたはSSHを使用してリポジトリをクローンできます。SSHを使用する場合は、事前にGitHubにSSH鍵を登録しておく必要があります。

```bash
# HTTPSを使用する場合
git clone https://github.com/username/repository.git

# SSHを使用する場合
git clone git@github.com:username/repository.git

# クローンしたディレクトリに移動
cd repository
```

## 基本的な開発フロー

### 1. リモートの最新化

新しい開発を始める前に、必ずローカルの状態を最新にします。これにより、他の開発者による変更との競合を防ぐことができます。

```bash
# mainブランチに切り替え
git checkout main

# リモートの変更を取得
git fetch origin

# ローカルのmainブランチを最新に更新
git pull origin main
```

### 2. 新規ブランチの作成

機能開発やバグ修正は、必ず新しいブランチを作成して行います。これにより、複数の開発を並行して進めることができ、また変更の管理が容易になります。

```bash
# 新しいブランチを作成して切り替え
git checkout -b feature/new-feature

# または
git branch feature/new-feature
git checkout feature/new-feature
```

ブランチ名は作業内容を反映した明確な名前をつけることが重要です。例えば:

- 機能追加: `feature/add-login`
- バグ修正: `fix/login-error`
- ドキュメント: `docs/api-reference`

### 3. 開発作業

開発中は定期的に変更の状態を確認し、適切な単位でコミットすることが重要です。

```bash
# ファイルの変更状態を確認
git status

# 変更をステージング
git add <ファイル名>
# または全ての変更をステージング
git add .

# 変更をコミット
git commit -m "feat: 新機能を追加"
```

### 4. リモートへのプッシュ

ローカルでの作業が完了したら、変更をリモートリポジトリにプッシュします。初回のプッシュ時は`-u`オプションを使用して、ローカルブランチとリモートブランチを紐付けます。

```bash
# 新規ブランチをリモートにプッシュ
git push -u origin feature/new-feature

# 2回目以降のプッシュ
git push origin feature/new-feature
```

### 5. プルリクエスト(PR)

プルリクエストは、開発した機能をメインブランチにマージするための仕組みです。以下の手順で作成します:

1. GitHubのリポジトリページで「Pull requests」タブを選択
2. 「New pull request」ボタンをクリック
3. ベースブランチ(通常はmain)と作業ブランチを選択
4. PRの内容を記述して提出

PRの説明には以下の内容を含めることを推奨します:

- 変更内容の概要
- 変更理由
- テスト方法
- 関連する課題(Issue)へのリンク

### 6. レビュー後のマージ

レビュープロセスを通じて、コードの品質を確保します:

1. レビューのコメントに対応して必要な修正を行う
2. 承認後、「Merge pull request」でマージ
3. ブランチの削除(オプション)

## コンフリクトの解決

### リモートの変更との競合

複数人で同じコードを編集している場合、コンフリクト(競合)が発生する可能性があります。以下の手順で解決します:

```bash
# 最新の変更を取得
git fetch origin

# リモートの変更を取り込む
git pull origin main

# コンフリクトが発生した場合
# 1. コンフリクトファイルを確認
git status

# 2. コンフリクトを修正
# 3. 修正後、変更をコミット
git add .
git commit -m "fix: コンフリクトを解決"
```

コンフリクトの解決時は、他の開発者と相談しながら慎重に進めることが重要です。

## 便利なコマンド

### 作業の一時退避

急な割り込み作業が発生した場合など、現在の作業を一時的に退避させることができます。

```bash
# 作業中の変更を一時保存
git stash save "作業メモ"

# 保存した作業の一覧表示
git stash list

# 保存した作業を復元
git stash pop
```

### ブランチ操作

ブランチの管理と操作のための基本的なコマンドです。

```bash
# ローカルブランチの一覧表示
git branch

# リモートブランチを含むすべてのブランチを表示
git branch -a

# ブランチの削除
git branch -d feature/old-feature
```

### コミット履歴の確認

プロジェクトの変更履歴を確認するための様々な方法があります。

```bash
# コミット履歴の表示
git log

# 簡潔な履歴表示
git log --oneline

# ブランチの視覚的な表示
git log --graph --oneline --all
```

## コミットメッセージの規約

コミットメッセージは、変更内容を明確に伝えるために重要です。以下の形式で記述することを推奨します:

```txt
<type>: <description>

[optional body]
[optional footer]
```

主なtype:

- feat: 新機能
- fix: バグ修正
- docs: ドキュメントのみの変更
- style: コードの動作に影響しない変更(フォーマット等)
- refactor: リファクタリング
- test: テストコードの追加・修正
- chore: ビルドプロセスや補助ツールの変更

良いコミットメッセージの例:

- `feat: ユーザー認証機能を追加`
- `fix: ログイン時のパスワード検証バグを修正`
- `docs: APIドキュメントを更新`

## 注意事項

- 定期的に `git pull` でリモートの変更を取り込む
- 大きな変更は小さな単位に分割してコミット
- コミットメッセージは具体的に記述
- 作業前に必ず最新のmainブランチから新しいブランチを作成
- 機密情報(パスワード、APIキーなど)をコミットしない
- 不明な場合はチームメンバーに相談
